#!/bin/bash

{{/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/}}

set -ex

function cleanup_network_trunks()
{
    TRUNKS=$(openstack network trunk list -c ID -f value)
    PORTS=$(openstack network trunk list -c "Parent Port" -f value)

    for TRUNK in ${TRUNKS}; do
        openstack network trunk delete ${TRUNK}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete trunk ${TRUNK}"
            return ${RET}
        fi
    done

    for PORT in ${PORTS}; do
        openstack port delete ${PORT}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete port ${PORT}"
            return ${RET}
        fi
    done
    return 0
}

function cleanup_vm_instances()
{
    local VMLIST=""
    local ID=""
    local RETRY=0

    VMLIST=$(openstack server list --all-projects -c ID -f value)
    for VM in ${VMLIST}; do
        openstack server delete ${VM} --wait
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete VM ${ID}"
            return ${RET}
        fi
    done

    return 0
}

function cleanup_floating_ips()
{
    local IPLIST=""
    local IP=""

    IPLIST=$(openstack floating ip list | grep -E "[0-9]+.[0-9]+.[0-9]+.[0-9]" | awk '{ print $2;  }')
    for IP in ${IPLIST}; do
        openstack floating ip delete ${IP}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete floating ip ${IP}"
            return 1
        fi
    done

    return 0
}

function cleanup_manual_ports()
{
    PORTS=$(openstack port list --device-owner=compute:manual | grep -E "^\|\s\w{8}-\w{4}-\w{4}-\w{4}-\w{12}\s\|" | awk '{ print $2;  }')
    for PORT in ${PORTS}; do
        openstack port delete ${PORT}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete manual port ${PORT}"
            return 1
        fi
    done

    return 0
}

function cleanup_routers()
{
    local ROUTERLIST=""
    local ID=""

    ROUTERLIST=$(openstack router list -c ID -f value)
    for ID in ${ROUTERLIST}; do
        openstack router set ${ID} --no-route
        openstack router unset --external-gateway ${ID}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to clear gateway on router ${ID}"
            return 1
        fi

        PORTS=$(openstack port list --router ${ID} -c ID -f value)
        for PORT in ${PORTS}; do
            openstack router remove port ${ID} ${PORT}
            RET=$?
            if [ ${RET} -ne 0  ]; then
                echo "Failed to delete interface ${PORT} from router ${ID}"
                return ${RET}
            fi
        done

        openstack router delete ${ID}
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete router ${ID}"
            return 1
        fi
    done

    return 0
}

function cleanup_application_ports()
{
    NETS=$(openstack network list -c ID -f value)
    for NET in $NETS; do
        NET_PORTS=$(openstack port list --network $NET -c ID -f value)
        for NET_PORT in $NET_PORTS; do
            openstack port delete $NET_PORT
            RET=$?
            if [ ${RET} -ne 0  ]; then
                echo "Failed to delete port ${NET_PORT}"
                return 1
            fi
        done
    done

    return 0
}

function cleanup_networks()
{
    local ID=""
    NETLIST=$(openstack network list -c ID -f value)
    for ID in ${NETLIST}; do
        openstack network delete ${ID}
        RET=$?
        if [ ${RET} -ne 0  ]; then
            echo "Failed to delete network ${ID}"
            return 1
        fi
    done

    return 0
}

date
echo "Cleaning up network resources..."

echo "Cleaning up network trunks"
cleanup_network_trunks
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup network trunks"
fi

echo "Cleaning up VM instances"
cleanup_vm_instances
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup VM instances"
fi

echo "Cleaning up floating IP addresses"
cleanup_floating_ips
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup floating IP addresses"
fi

echo "Cleaning up manual ports"
cleanup_manual_ports
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup manual ports"
fi

echo "Cleaning up routers"
cleanup_routers
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup routers"
fi

echo "Cleaning up application ports"
cleanup_application_ports
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup shared networks"
fi

echo "Cleaning up networks"
cleanup_networks
RET=$?
if [ ${RET} -ne 0  ]; then
    echo "Failed to cleanup networks"
fi

date
echo "Cleanup finished"

exit 0
